/*
                ##### RASPBERRY PI <==> PROPELLER COMMUNICATIONS #####
   This test software is used in conjunction with RpiTest.spin which runs on the
   Parallax Propeller microprocessor.  It is intended to implement communications
   between the Raspberry Pi and the Propeller


*/
#include <bcm2835.h>
#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

//  GPIO Pin Assignments
#define GndPin   RPI_GPIO_P1_06   // Ground
// ****** << Led Assignments >> *****
#define Red      RPI_GPIO_P1_11   // Red    Led
#define Yellow   RPI_GPIO_P1_13   // Yellow Led
#define Green    RPI_GPIO_P1_15   // Green  Led
// ****** << I2C Assignments >> *****
#define SDA      RPI_GPIO_P1_03   // DATA
#define SCL      RPI_GPIO_P1_05   // CLOCK
// ****** << SPI Assignments >> *****
#define ClkPin   RPI_GPIO_P1_23   // Clock Pin
#define MISOPin  RPI_GPIO_P1_21   // Master In Slave Out Pin
#define MOSIPin  RPI_GPIO_P1_19   // Master Out Slave In Pin 
#define Csel0    RPI_GPIO_P1_24   // Chip Select 0
#define Csel1    RPI_GPIO_P1_26   // Chip Select 1

// ======<< Ascii control Characters >>=====
#define      soh  0x01
#define      stx  0x02
#define      etx  0x03
#define      eot  0x04
#define      ack  0x06
#define      nak  0x15
#define      syn  0x16

#define      Speed  10000                //Hz 
#define      SpiClk 1000/Speed

typedef union { unsigned int MyLong; unsigned char MyByte[4]; } LongByte;
unsigned char SndPkt[100], RcvPkt[100];

/*******************************
 **                           **
 **   Pin Control Routines    **
 **                           **
 *******************************/
void Pin(int Pin, int level) { bcm2835_gpio_write(Pin, level); }
int ReadPin(int Pin) { return bcm2835_gpio_lev(Pin); }


/*********************************************************
 **                      SpiRead Byte                   **
 **                                                     **
 **  This routine receives one Unit of data             **
 **  one bit at a time as follows:                      **
 **  repeat for N bits                                  **
 **       Wait while the clock Pin is LOW               **
 **       Set input byte bit 0 equal to the MISO Pin    **
 **       Shift input byte one place left               **
 **       Wait while the clock Pin is HIGH              **
 **                                                     **
 *********************************************************/
int SpiRead(int N)
{       
    unsigned char bit;
    unsigned int indat;

    indat = 0;
    for (bit=0; bit<N; bit++) {
        Pin(ClkPin,LOW);
        delay(SpiClk/2);
        Pin(ClkPin,HIGH);
        delay(SpiClk/2);
        indat <<=  1;
        indat |=  ReadPin(MISOPin);
    }
    Pin(ClkPin,LOW);
    return indat;
}


/********************************************************
 **                 SpiWrite Byte                      ** 
 **                                                    **
 **  This routine sends one byte of data               **
 **  one bit at a time as follows:                     **
 **  repeat for 8 bits                                 **
 **       Set the clock PIn LOW                        **
 **       Set MOSI Pin equal to bit 8 of the data byte **
 **       Shift the output byte one bit left           **
 **       wait one half clock period                   **
 **       Set the clock Pin HIGH                       **
 **       wait one half clock period                   **
 **                                                    **
 ********************************************************/
void SpiWrite(unsigned char ch, int N)
{       
    unsigned char bit;

    Pin(ClkPin, LOW);
    for (bit=0; bit<N; bit++) {
        //  Send bit N (the MSB) and shift left one place 
        bcm2835_gpio_write(MOSIPin, (ch & 0x80) );
        ch <<= 1;   
        delay(SpiClk/2);
        Pin(ClkPin, HIGH);
        delay(SpiClk/2);
        Pin(ClkPin, LOW);
    }
}

/********************************************************
 **                                                    **
 **          Utility Pack / UnPack  Routines           ** 
 **                                                    **
 **  These routines add/remove numbers from the Packet **
 **                                                    **
 ********************************************************/
void PackLong( char* p, LongByte N ) {
    *p = N.MyByte[0];  p++;
    *p = N.MyByte[1];  p++;
    *p = N.MyByte[2];  p++;
    *p = N.MyByte[3];  p++;
}

LongByte UnPackLong( char* p ) {
    LongByte N; 

    N.MyByte[0] = *p;  p++;
    N.MyByte[1] = *p;  p++;
    N.MyByte[2] = *p;  p++;
    N.MyByte[3] = *p;  p++;
    return N;
}

/**************************************************************************************
 **                                                                                  **
 **                      Send / Receive Packet  Routines                             ** 
 **                   These routines Send or receive a Packet                        **
 **                                                                                  **
 ** -------------------------------------------------------------------------------- **
 **                      Packet Definition                                           **
 **     Each block begins with <soh> <stx> and a <count> followed by the data        **
 **     bytes and the <checksum>. There are no restrictions on the contents of       **
 **     The Packet. The checksum is the mod 256 sum of the data not including        **
 **     soh,stx,count or cksum, count does not include soh,stx,count or cksum.       **
 **     <packet>      ::= <packetstart><count><data><cksum>                          **
 **     <packetstart> ::= <headerstart><textstart>                                   **
 **     <headerstart> ::= soh                                                        **
 **     <textstart>   ::= stx                                                        **
 **     <count>       ::= <maxnumber>                                                **
 **     <cksum>       ::= Sum<data> mod 256                                          **
 **     <data>        ::= <byte> | <data><byte                                       **
 **     <byte>        ::= $00..$FF                                                   **
 **     <maxnumber>   ::= 0..255                                                     **
 **                                                                                  **
 **************************************************************************************/
/***************************************
 **          SEND PACKET              **
 ***************************************/
void SndPacket( int N) {
   unsigned char  i, Ck1;
   
   Ck1 = 0;
   SndPkt[0] = stx;                // Init start of packet
   Pin(Csel0, LOW);                // Select Low
   delay(SpiClk/2);
   SpiWrite(N,8);                  // Send Packet Size
   for ( i=0; i<N; i++ ) { SpiWrite(SndPkt[i] ,8); Ck1 += (SndPkt[i] % 256); } 
   SpiWrite(Ck1,8);                // Send the checksum
   delay(SpiClk/2);
   Pin(Csel0, HIGH);              // Select high
}

/************************************
 **       RECEIVE PACKET           **
 ************************************/
int RcvPacket() {
   unsigned char  i, N, Ck1, Ck2;
  
   Ck1 = 0;                        // Start with both checksums at zero
   Ck2 = 0;
   Pin(Csel0, LOW);                // Select Low
   delay(SpiClk/2);
   N = SpiRead(8);
   for (i=0; i<N; i++) { RcvPkt[i] = SpiRead(8); Ck1 += (RcvPkt[i] % 256); }
   Ck2 = SpiRead(8);              // Get the senders checksum
   delay(SpiClk/2);
   Pin(Csel0, HIGH);              // Select high
   return N;
   if (Ck1==Ck2) { return N; } else { return 0; }
  
}

void FillPkt( int N) {
   int i;
   for (i=0; i<N; i++) { SndPkt[i] = (unsigned char)(i+0x30); }
}

void ShowRcvPkt(int N) {
   int i;
   for (i=0; i<N; i++) { printf("%2x ",RcvPkt[i]); }
   printf("\n");
}

/*********************
 **      MAIN       **
 *********************/
int main(int argc, char **argv)
{
    int i, j, PktSize;
    LongByte MyData;
    char line1[] = "0123456789.ABCDEFGHIJ.0123456789."; // this is a test line
    unsigned char* p1;

    if (!bcm2835_init()) { return 1; }

    // Set the output pins
// ****** << I2C Assignments >> *****
    bcm2835_gpio_fsel(SCL, BCM2835_GPIO_FSEL_OUTP);        // Set SCL to be output
    bcm2835_gpio_fsel(SDA, BCM2835_GPIO_FSEL_OUTP);        // Make SDA pin an output
// ****** << Spi Assignments >> *****
    bcm2835_gpio_fsel(ClkPin,   BCM2835_GPIO_FSEL_OUTP);   // ClkPin is an output;
    bcm2835_gpio_fsel(MOSIPin,  BCM2835_GPIO_FSEL_OUTP);   // MOSI is an output;
    bcm2835_gpio_fsel(MISOPin,  BCM2835_GPIO_FSEL_INPT);   // MISO is an input
//    bcm2835_gpio_set_pud(MISOPin, BCM2835_GPIO_PUD_UP);    // with a pullup
    bcm2835_gpio_fsel(Csel0,    BCM2835_GPIO_FSEL_OUTP);   // CSel0 is an output
    bcm2835_gpio_fsel(Csel1,    BCM2835_GPIO_FSEL_OUTP);   // CSel0 is an output;


// Create a test long and pack it into the test line
// The propeller will receive it, display it, replace it with 202202, and send it back    
      
    Pin(Csel0,HIGH);
    Pin(ClkPin,LOW);
    j = 0;
    PktSize = 33;
    while (1) {
    
       FillPkt(PktSize);
       PktSize = 33;  
       p1 = &SndPkt[1];                // Address of test long in the packet
       MyData.MyLong = j++;            // create a test long
       PackLong(p1, MyData);           // Pack it into the test line
       SndPacket(PktSize);   

       delay(50);                      // Give the Propeller time to process / Display
    
       PktSize = RcvPacket();
    
       p1 = &RcvPkt[1];                // p1 gets address of long in the packet
       MyData = UnPackLong(p1); 
       printf("A:%d  ", MyData.MyLong, 4);
       p1 = &RcvPkt[5];                // p1 gets address of long in the packet
       MyData = UnPackLong(p1); 
       printf("B:%d  ", MyData.MyLong, 4);
  
       printf("Recieved: ");           // Display returned data
       p1 = &RcvPkt[0];
       ShowRcvPkt(PktSize);
       delay(50);                      // Slow things down so humans can observe
      
    }
}
